@@ -48,12 +48,35 @@
 (define %cache-delay (make-parameter 0))
 
 (define (workflow->data-hashes workflow processes free-inputs-map make-script)
-  "Return an alist associating each of the WORKFLOW's PROCESSES with
-the hash of all the process scripts used to generate their outputs.
-FREE-INPUTS-MAP is an alist of input names to file names that must be
-considered when computing the hash."
+  "Return an alist associating each of the WORKFLOW's ordered list of
+PROCESSES with the hash of all the process scripts used to generate
+their outputs.  FREE-INPUTS-MAP is an alist of input names to file
+names that must be considered when computing the hash."
   (define graph (workflow-restrictions workflow))
+  (define (process-free-inputs process)
+    (filter-map (lambda (input)
+                  (and=> (assoc-ref free-inputs-map input) first))
+                (process-inputs process)))
+  ;; XXX: We don't hash all of the inputs, because that might be very
+  ;; expensive --- especially when done *every* time a workflow is
+  ;; supposed to be run.  We just hash the file name, the mtime and
+  ;; the file size and hope that's enough.
 
+  ;; XXX: We should probably have a cache for file hashes, so that
+  ;; hashing even large files can be acceptable as it only has to be
+  ;; done once.
+  (define input-hashes
+    (map (match-lambda
+           ((name file-name)
+            (let ((st (stat file-name)))
+              (cons file-name
+                    (bytevector->u8-list
+                     (sha256 (string->bytevector (format #f "~a~a~a"
+                                                         file-name
+                                                         (stat:mtime st)
+                                                         (stat:size st))
+                                                 "ISO-8859-1")))))))
+         free-inputs-map))
   ;; Compute hashes for chains of scripts.
   (define (kons process acc)
     (let* ((script (make-script process #:workflow workflow))
@@ -63,46 +86,19 @@ considered when computing the hash."
        (cons process
              (append hash
                      ;; Hash of mapped free inputs.
-                     (match (filter-map (lambda (input)
-                                          (and=> (assoc-ref free-inputs-map input) first))
-                                        (process-inputs process))
-                       (() (list))
-                       (file-names
-                        (append-map (lambda (file-name)
-                                      (assoc-ref acc file-name))
-                                    file-names)))
+                     (append-map (cut assoc-ref acc <>)
+                                 (process-free-inputs process))
                      ;; All outputs of processes this one depends on.
                      (append-map (cut assoc-ref acc <>)
                                  (or (assoc-ref graph process) '()))))
        acc)))
-
-  ;; XXX: We don't hash all of the inputs, because that might be very
-  ;; expensive --- especially when done *every* time a workflow is
-  ;; supposed to be run.  We just hash the file name, the mtime and
-  ;; the file size and hope that's enough.
-
-  ;; XXX: We should probably have a cache for file hashes, so that
-  ;; hashing even large files can be acceptable as it only has to be
-  ;; done once.
-  (let ((input-hashes
-         (map (match-lambda
-                ((name file-name)
-                 (let ((st (stat file-name)))
-                   (cons file-name
-                         (bytevector->u8-list
-                          (sha256 (string->bytevector (format #f "~a~a~a"
-                                                              file-name
-                                                              (stat:mtime st)
-                                                              (stat:size st))
-                                                      "ISO-8859-1")))))))
-              free-inputs-map)))
-    (map (match-lambda
-           ((process . hashes)
-            (cons process
-                  (bytevector->base32-string
-                   (sha256
-                    (u8-list->bytevector hashes))))))
-         (fold kons input-hashes processes))))
+  (map (match-lambda
+         ((process . hashes)
+          (cons process
+                (bytevector->base32-string
+                 (sha256
+                  (u8-list->bytevector hashes))))))
+       (fold kons input-hashes processes)))
 
 (define (make-process->cache-prefix workflow free-inputs-map ordered-processes
                                     make-script)
@@ -126,7 +122,7 @@ prefix for its outputs."
 
 (define (cache! file cache-prefix)
   "Cache FILE by linking it to the directory CACHE-PREFIX."
-  (unless directory?
+  (unless (directory? file)
     (mkdir-p (string-append cache-prefix (dirname file)))
     ;; TODO: cross-device links don't work, of course.  Copy?  Symlink?
     ;; Make this configurable?
@@ -137,7 +133,7 @@ prefix for its outputs."
 
 (define (restore! file cache-prefix)
   "Restore FILE from the cache at CACHE-PREFIX."
-  (unless directory?
+  (unless (directory? file)
     (mkdir-p (dirname file))
     (when (file-exists? file)
       (delete-file file))
