@@ -27,6 +27,8 @@
   #:use-module (srfi srfi-9 gnu)
   #:use-module (srfi srfi-11)
   #:use-module (srfi srfi-26)
+  #:use-module (srfi srfi-34)
+  #:use-module (srfi srfi-35)
   #:use-module (oop goops)
   #:export (make-workflow
             workflow?
@@ -269,6 +271,65 @@ can be used in a fold over the WORKFLOW's processes."
                             (process->script engine #:containerize? containerize?))
                            '() ordered-processes))))
 
+(define (inputs->map inputs)
+  "Given a list of strings INPUTS of the format \"a=b\" or just \"a\",
+return a normalized mapping as a list of two element lists containing
+\"a\" and \"b\" or just \"a\" and \"a\"."
+  (map (lambda (value)
+         ;; A mapping is optional, so normalize it.
+         (if (string-contains value "=")
+             (string-split value #\=)
+             (list value value)))
+       inputs))
+
+(define-condition-type &missing-inputs &condition
+  missing-inputs-condition?
+  (files missing-inputs-files))
+
+(define (prepare-inputs workflow inputs-map)
+  "Ensure that all files in the INPUTS-MAP alist exist and are linked
+to the expected locations.  Pick unspecified inputs from the
+environment.  Return either the INPUTS-MAP alist with any additionally
+used input file names added, or raise a condition containing the list
+of missing files."
+  (define-values (input-names input-files)
+    (match inputs-map
+      (() (values '() '()))
+      (_ (apply values
+                (apply zip inputs-map)))))
+  (define unspecified-inputs
+    (lset-difference equal?
+                     (workflow-free-inputs workflow)
+                     input-names))
+
+  ;; TODO: can these two cases be merged?
+  (match unspecified-inputs
+    (()
+     ;; verify input files
+     (match (filter (negate file-exists?) input-files)
+       (()
+        ;; Link all mapped input files to their target locations
+        ;; TODO: ensure that target directories exist.
+        (for-each (match-lambda
+                    ((target source)
+                     (unless (file-exists? target)
+                       (link source target))))
+                  inputs-map)
+        inputs-map)
+       (missing
+        (raise (condition
+                (&missing-inputs (files missing)))))))
+    (missing
+     ;; Try to find the files in the environment.
+     (let* ((found (filter file-exists? missing))
+            (really-missing (lset-difference equal? missing found)))
+       (if (null? really-missing)
+           (append inputs-map
+                   (map (lambda (file) (list file file))
+                        found))
+           (raise (condition
+                   (&missing-inputs (files really-missing)))))))))
+
 (define* (workflow-run workflow engine
                        #:key
                        (inputs '())
@@ -284,150 +345,97 @@ to existing files.
 
 When CONTAINERIZE? is #T build a process script that spawns a
 container."
-  (define inputs-map
-    (match (map (lambda (value)
-                  ;; A mapping is optional, so normalize it.
-                  (if (string-contains value "=")
-                      (string-split value #\=)
-                      (list value value)))
-                inputs)
-      (() '())
-      (mapping mapping)))
-
-  (define-values (input-names input-files)
-    (match inputs-map
-      (() (values '() '()))
-      (_ (apply values
-                (apply zip inputs-map)))))
-
-  (define unspecified-inputs
-    (lset-difference equal?
-                     (workflow-free-inputs workflow)
-                     input-names))
-
-  (define (inputs-valid?)
-    (match unspecified-inputs
-      (()
-       ;; verify input files
-       (match (filter (negate file-exists?) input-files)
-         (()
-          ;; Link all mapped input files to their target locations
-          ;; TODO: ensure that target directories exist.
-          (unless (null? inputs-map)
-            (for-each (match-lambda
-                        ((target source)
-                         (unless (file-exists? target)
-                           (link source target))))
-                      inputs-map))
-          #t)
-         (missing
-          (format (current-error-port)
-                  "Missing files: ~{~%  * ~a~}.~%"
-                  missing)
-          #f)))
-      (missing
-       ;; Try to find the files in the environment.
-       ;; XXX Tell user that we pick the files from the current
-       ;; working directory.
-       ;; XXX These files would need to be mapped into the
-       ;; container.
-       (let* ((found (filter file-exists? missing))
-              (really-missing (lset-difference equal? missing found)))
-         (if (null? really-missing)
-             found
-             (begin (format (current-error-port)
-                            "Missing inputs: ~{~%  * ~a~}.~%Provide them with --input=NAME=FILE.~%"
-                            really-missing)
-                    #f))))))
+  (define inputs-map (inputs->map inputs))
+  (define inputs-map-with-extra-files
+    (guard (condition
+            ((missing-inputs-condition? condition)
+             (format (current-error-port)
+                     "Missing inputs: ~{~%  * ~a~}.~%Provide them with --input=NAME=FILE.~%"
+                     (missing-inputs-files condition))
+             (exit 1)))
+      (prepare-inputs workflow inputs-map)))
   (define ordered-processes
     (workflow-run-order workflow #:parallel? parallel?))
-  (define (run input-files)
-    (let ((make-script (process->script engine #:containerize? containerize?))
-          (runner (process-engine-runner engine)))
-      (define process->cache-prefix
-        (make-process->cache-prefix workflow
-                                    (append inputs-map
-                                            ;; Consider changes to
-                                            ;; unspecified inputs that
-                                            ;; are picked up from the
-                                            ;; current working
-                                            ;; directory.
-                                            (map (lambda (x) (list x x))
-                                                 unspecified-inputs))
-                                    ordered-processes
-                                    make-script))
-      (define cached?
-        (if force?
-            (const #f)
-            (lambda (process)
-              (and (not (null? (process-outputs process)))
-                   (let ((cache-prefix (process->cache-prefix process)))
-                     (every (lambda (out)
-                              (file-exists?
-                               (string-append cache-prefix out)))
-                            (process-outputs process)))))))
-      (lambda* (process #:key workflow)
-        (let ((cache-prefix (process->cache-prefix process)))
-          (if (cached? process)
-              (if dry-run?
+  (define make-script
+    (process->script engine #:containerize? containerize?))
+  (define runner (process-engine-runner engine))
+  (define process->cache-prefix
+    (make-process->cache-prefix workflow
+                                inputs-map-with-extra-files
+                                ordered-processes
+                                make-script))
+  (define cached?
+    (if force?
+        (const #f)
+        (lambda (process)
+          (and (not (null? (process-outputs process)))
+               (let ((cache-prefix (process->cache-prefix process)))
+                 (every (lambda (out)
+                          (file-exists?
+                           (string-append cache-prefix out)))
+                        (process-outputs process)))))))
+  (define* (run process #:key workflow)
+    (let ((cache-prefix (process->cache-prefix process)))
+      (if (cached? process)
+          (if dry-run?
+              (format (current-error-port)
+                      "Would skip process \"~a\" (cached at ~a).~%"
+                      (process-name process)
+                      cache-prefix)
+              (begin
+                (format (current-error-port)
+                        "Skipping process \"~a\" (cached at ~a).~%"
+                        (process-name process)
+                        cache-prefix)
+                ;; TODO: mount the cache directory in the container
+                ;; if containerized.  Otherwise link files from
+                ;; cache to expected location.
+                (for-each (cut restore! <> cache-prefix)
+                          (process-outputs process))))
+
+          ;; Not cached: execute the process!
+          (let* ((script
+                  (make-script process
+                               #:workflow workflow
+                               #:input-files
+                               (lset-intersection
+                                string=?
+                                (map second inputs-map-with-extra-files)
+                                (process-inputs process))))
+                 (command (append runner (list script))))
+            (if dry-run?
+                (format (current-error-port)
+                        "Would execute: ~{~a ~}~%" command)
+                (begin
                   (format (current-error-port)
-                          "Would skip process \"~a\" (cached at ~a).~%"
-                          (process-name process)
-                          cache-prefix)
-                  (begin
-                    (format (current-error-port)
-                            "Skipping process \"~a\" (cached at ~a).~%"
-                            (process-name process)
-                            cache-prefix)
-                    ;; TODO: mount the cache directory in the container
-                    ;; if containerized.  Otherwise link files from
-                    ;; cache to expected location.
-                    (for-each (cut restore! <> cache-prefix)
-                              (process-outputs process))))
-
-              ;; Not cached: execute the process!
-              (let ((command (append runner
-                                     (list
-                                      (make-script
-                                       process
-                                       #:workflow workflow
-                                       #:input-files
-                                       (lset-intersection
-                                        string=?
-                                        input-files
-                                        (process-inputs process)))))))
-                (if dry-run?
-                    (format (current-error-port)
-                            "Would execute: ~{~a ~}~%" command)
-                    (begin
+                          "Executing: ~{~a ~}~%" command)
+                  (let ((retval (apply system* command)))
+                    (unless (zero? retval)
                       (format (current-error-port)
-                              "Executing: ~{~a ~}~%" command)
-                      (let ((retval (apply system* command)))
-                        (unless (zero? retval)
-                          (format (current-error-port)
-                                  "error: process `~a' failed with return value ~a.~%"
-                                  (process-name process)
-                                  retval)
-                          (exit retval)))
-                      ;; Wait before generated files are accessed.
-                      ;; This may be needed for distributed file
-                      ;; systems.
-                      (usleep (%cache-delay))
+                              "error: process `~a' failed with return value ~a.~%"
+                              (process-name process)
+                              retval)
+                      (exit retval)))
+                  ;; Wait before generated files are accessed.
+                  ;; This may be needed for distributed file
+                  ;; systems.
+                  (usleep (%cache-delay))
 
-                      ;; Abort if declared outputs are missing.
-                      (for-each (lambda (output)
-                                  (unless (file-exists? (string-append (getcwd) "/" output))
-                                    (format (current-error-port)
-                                            "error: process `~a' failed to produce output ~a.~%"
-                                            (process-name process)
-                                            output)
-                                    (exit 1)))
-                                (process-outputs process))
+                  ;; Abort if declared outputs are missing.
+                  (for-each (lambda (output)
+                              (let ((canonical-name (if (absolute-file-name? output)
+                                                        output
+                                                        (string-append (getcwd) "/" output))))
+                                (unless (file-exists? canonical-name)
+                                  (format (current-error-port)
+                                          "error: process `~a' failed to produce output ~a.~%"
+                                          (process-name process)
+                                          output)
+                                  (exit 1))))
+                            (process-outputs process))
 
-                      ;; Link files to the cache.
-                      (for-each (cut cache! <> cache-prefix)
-                                (process-outputs process))))))))))
-  (and=> (inputs-valid?)
-         (lambda (input-files)
-           (fold (workflow-kons workflow (run input-files))
-                 '() ordered-processes))))
+                  ;; Link files to the cache.
+                  (for-each (cut cache! <> cache-prefix)
+                            (process-outputs process))))))))
+  (fold (workflow-kons workflow run)
+        '() ordered-processes))
